"""
interpolate - libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
from libscientific.loadlibrary import load_libscientific_library
import libscientific.matrix as mx
import libscientific.vector as vect

lsci = load_libscientific_library()


lsci.cubic_spline_interpolation.argtypes = [ctypes.POINTER(mx.MATRIX),
                                            ctypes.POINTER(mx.MATRIX)]
lsci.cubic_spline_interpolation.restype = None


def cubic_spline_interpolation(x_y, s_interp):
    """
    CubicSplineInterpolation: Given a matrix x_y, interpolate the npoints
                              using a natural cubic spline approach
    """
    x_y_ = None
    if isinstance(x_y, mx.Matrix):
        x_y_ = x_y.mx
    else:
        x_y_ = x_y

    s_interp_ = None
    if isinstance(S, mx.Matrix):
        s_interp_ = s_interp.mx
    else:
        s_interp_ = s_interp

    lsci.cubic_spline_interpolation(x_y_, s_interp_)


lsci.cubic_spline_predict.argtypes = [ctypes.POINTER(vect.DVECTOR),
                                      ctypes.POINTER(mx.MATRIX),
                                      ctypes.POINTER(vect.DVECTOR)]
lsci.cubic_spline_predict.restype = None


def cubic_spline_predict(x_cc, s_interp, y_pred):
    """
    Natural cubic spline prediction
    """
    x_cc_ = None
    if isinstance(x, vect.DVector):
        x_cc_ = x_cc.dvect
    else:
        x_cc_ = x_cc

    s_interp_ = None
    if isinstance(s_interp, mx.Matrix):
        s_interp_ = s_interp.mx
    else:
        s_interp_ = s_interp

    y_pred_ = None
    if isinstance(y_pred, vect.DVector):
        y_pred_ = y_pred.d
    else:
        y_pred_ = y_pred

    lsci.cubic_spline_predict(x_cc_, s_interp_, y_pred_)


if __name__ in "__main__":
    xy = [[-1.82,0.63],
           [-0.73,0.19],
           [-0.17,0.01],
           [-0.09,0.00],
           [0.15,0.01],
           [0.39,0.06],
           [0.86,0.24],
           [1.44,0.49]]
    x = [-1.82, -0.73, -0.17, -0.09, 0.15, 0.39, 0.86, 1.44]
    xy = mx.new_matrix(xy)
    print("Origin matrix x->y to interpolate")
    mx.print_matrix(xy)
    S = mx.init_matrix()
    cubic_spline_interpolation(xy, S)
    print("Spline interpolation coefficient matrix")
    mx.print_matrix(S)
    x = vect.DVector(x)
    print("Vector to predict")
    x.debug()
    yp = vect.init_dvector()
    cubic_spline_predict(x, S, yp)

    print("Prediction")
    vect.print_dvector(yp)
    mx.del_matrix(S)
    vect.del_dvector(yp)
    mx.del_matrix(xy)



"""
pca libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
import libscientific.matrix as mx
import libscientific.vector as vect
from libscientific.loadlibrary import load_libscientific_library

lsci = load_libscientific_library()


class PCAMODEL(ctypes.Structure):
    """
    PCA model data structure
    """
    _fields_ = [
        ("scores", ctypes.POINTER(mx.MATRIX)),
        ("loadings", ctypes.POINTER(mx.MATRIX)),
        ("dmodx", ctypes.POINTER(mx.MATRIX)),
        ("varexp", ctypes.POINTER(vect.DVECTOR)),
        ("colaverage", ctypes.POINTER(vect.DVECTOR)),
        ("colscaling", ctypes.POINTER(vect.DVECTOR))]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__



lsci.NewPCAModel.argtypes = [ctypes.POINTER(ctypes.POINTER(PCAMODEL))]
lsci.NewPCAModel.restype = None


def new_pca_model():
    """
    new_pca_model: Allocate in memory an empty libscientific PCA model
    """
    mpca = ctypes.POINTER(PCAMODEL)()
    lsci.NewPCAModel(ctypes.pointer(mpca))
    return mpca


lsci.DelPCAModel.argtypes = [ctypes.POINTER(ctypes.POINTER(PCAMODEL))]
lsci.DelPCAModel.restype = None


def del_pca_model(mpca):
    """
    del_pca_model: Delete an allocated libscientific PCA model
    """
    lsci.DelPCAModel(ctypes.pointer(mpca))


lsci.PCA.argtypes = [ctypes.POINTER(mx.MATRIX),
                     ctypes.c_size_t,
                     ctypes.c_size_t,
                     ctypes.POINTER(PCAMODEL),
                     ctypes.POINTER(ctypes.c_int)]
lsci.PCA.restype = None


def pca_algorithm(m_input, scaling, npc, mpca):
    """
    PCA: Calculate the PCA model for a matrix m using the NIPALS algorithm
    """
    ssignal = ctypes.c_int(0)
    lsci.PCA(m_input, scaling, npc, mpca, ssignal)


lsci.PCAScorePredictor.argtypes = [ctypes.POINTER(mx.MATRIX),
                                   ctypes.POINTER(PCAMODEL),
                                   ctypes.c_size_t,
                                   ctypes.POINTER(mx.MATRIX)]
lsci.PCAScorePredictor.restype = None


def pca_score_predictor(m_input, mpca, npc, pscores):
    """
    pca_score_predictor: Predict scores for a matrix m in the computed PCA modes
    """
    lsci.PCAScorePredictor(m_input,
                           mpca,
                           npc,
                           pscores)


lsci.PCAIndVarPredictor.argtypes = [ctypes.POINTER(mx.MATRIX),
                                    ctypes.POINTER(mx.MATRIX),
                                    ctypes.POINTER(vect.DVECTOR),
                                    ctypes.POINTER(vect.DVECTOR),
                                    ctypes.c_size_t,
                                    ctypes.POINTER(mx.MATRIX)]
lsci.PCAIndVarPredictor.restype = None


def pca_ind_var_predictor(scores_input,
                          loadings_input,
                          colaverage,
                          colscaling,
                          npc,
                          ind_vars_out):
    """
    pca_ind_var_predictor: Reconstruct the original matrix
                           from PCA model using scores and loadings
    """
    lsci.PCAIndVarPredictor(scores_input,
                            loadings_input,
                            colaverage,
                            colscaling,
                            npc,
                            ind_vars_out)


lsci.PrintPCA.argtypes = [ctypes.POINTER(PCAMODEL)]
lsci.PrintPCA.restype = None


def print_pca(mpca):
    """
    PrintPCA: Print to video the PCA Model
    """
    lsci.PrintPCA(mpca)


class PCA():
    """
    PCA model class
    """
    def __init__(self, scaling, npc):
        self.mpca = new_pca_model()
        self.scaling = scaling
        self.npc = npc

    def __del__(self):
        if self.mpca is not None:
            del_pca_model(self.mpca)
            del self.mpca
        self.mpca = None

    def fit(self, m_input):
        """
        fit a PCA model using an input matrix
        """
        if "Matrix" in str(type(m_input)):
            pca_algorithm(m_input, self.scaling, self.npc, self.mpca)
        else:
            m_input_ = mx.new_matrix(m_input)
            pca_algorithm(m_input_, self.scaling, self.npc, self.mpca)
            mx.del_matrix(m_input_)
            del m_input_

    def get_scores(self):
        """
        get the PCA scores
        """
        return mx.matrix_to_list(self.mpca[0].scores)

    def get_loadings(self):
        """
        get the PCA loadings
        """
        return mx.matrix_to_list(self.mpca[0].loadings)

    def get_exp_variance(self):
        """
        get the PCA explained variance
        """
        return vect.dvector_tolist(self.mpca[0].varexp)

    def predict(self, m_input):
        """
        Project an input matrix into the PCA model
        """
        m_input_ = mx.new_matrix(m_input)
        pscores_ = mx.init_matrix()
        pca_score_predictor(m_input_, self.mpca, self.npc, pscores_)
        pscores = mx.matrix_to_list(pscores_)
        mx.del_matrix(m_input_)
        del m_input_
        mx.del_matrix(pscores_)
        del pscores_
        return pscores

    def reconstruct_original_matrix(self,
                                    npc_input=None,
                                    scores_input=None):
        """
        Reconstruct the original input matrix giving a
        number of principal components to be used from scores and loadings
        """
        npc_ = None
        if npc_input is None:
            npc_ = self.npc
        else:
            npc_ = npc_input

        scores_ = None
        if scores_input is None:
            scores_ = self.mpca[0].scores
        else:
            scores_ = scores_input

        ind_vars = mx.init_matrix()
        pca_ind_var_predictor(scores_,
                              self.mpca[0].loadings,
                              self.mpca[0].colaverage,
                              self.mpca[0].colscaling,
                              npc_,
                              ind_vars)
        omx = mx.matrix_to_list(ind_vars)
        mx.del_matrix(ind_vars)
        del ind_vars
        return omx


if __name__ == '__main__':
    def mx_to_video(m_input, decimals=5):
        """
        print a matrix to video
        """
        for row in m_input:
            print("\t".join([str(round(x, decimals)) for x in row]))
    import random
    random.seed(123456)
    a = [[random.random() for j in range(4)] for i in range(10)]
    print("Original Matrix")
    mx_to_video(a)
    print("Computing PCA ...")
    model = PCA(1, 2)
    model.fit(a)
    print("Showing the PCA scores")
    scores = model.get_scores()
    mx_to_video(scores, 3)
    print("Showing the PCA loadings")
    loadings = model.get_loadings()
    mx_to_video(loadings, 3)
    print(model.get_exp_variance())
    print("Reconstruct the original PCA matrix using the PCA Model")
    ra = model.reconstruct_original_matrix()
    mx_to_video(ra)
    pred_scores = model.predict(a)
    for i, row1 in enumerate(pred_scores):
        row2 = scores[i]
        print(row1, row2)



"""
vector libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
from libscientific.loadlibrary import load_libscientific_library
from libscientific import misc

lsci = load_libscientific_library()


class STRVECTOR(ctypes.Structure):
    """
    string vector data structure
    """
    _fields_ = [
        ("data", ctypes.POINTER(ctypes.POINTER(ctypes.c_char))),
        ("size", ctypes.c_size_t)]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__


class DVECTOR(ctypes.Structure):
    """
    double vector data structure
    """
    _fields_ = [
        ("data", ctypes.POINTER(ctypes.c_double)),
        ("size", ctypes.c_size_t)]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__


lsci.initDVector.argtypes = [ctypes.POINTER(ctypes.POINTER(DVECTOR))]
lsci.initDVector.restype = None


def init_dvector():
    """
    initDVector: Allocate in memory an empty libscientific double vector
    """
    d_vect = ctypes.POINTER(DVECTOR)()
    lsci.initDVector(ctypes.pointer(d_vect))
    return d_vect


lsci.NewDVector.argtypes = [ctypes.POINTER(ctypes.POINTER(DVECTOR)),
                            ctypes.c_size_t]
lsci.NewDVector.restype = None


def new_dvector(v_lst):
    """
    NewDVector: Allocate in memory a libscientific double vector from a list
    """
    size = len(v_lst)
    d_vect = ctypes.POINTER(DVECTOR)()
    lsci.NewDVector(ctypes.pointer(d_vect), size)

    for i in range(size):
        val = None
        try:
            val = float(v_lst[i])
        except ValueError:
            val = None
        if val is None:
            lsci.setDVectorValue(d_vect, i, misc.missing_value())
        else:
            lsci.setDVectorValue(d_vect, i, val)
    return d_vect

lsci.DelDVector.argtypes = [ctypes.POINTER(ctypes.POINTER(DVECTOR))]
lsci.DelDVector.restype = None

def del_dvector(dvect):
    """
    DelDVector: Delete an allocated libscientific double vector
    """
    lsci.initDVector(ctypes.pointer(dvect))

lsci.DVectorResize.argtypes = [ctypes.POINTER(DVECTOR),
                               ctypes.c_size_t]
lsci.DVectorResize.restype = None


def dvector_resize(dvect, size):
    """
    DVectorResize: Resize an already allocated or reallocate a libscientific
                   double vector
    """
    lsci.DVectorResize(dvect, size)


lsci.PrintDVector.argtypes = [ctypes.POINTER(DVECTOR)]
lsci.PrintDVector.restype = None


def print_dvector(dvect):
    """
    PrintDVector: Print to video a libscientific double vector
    """
    lsci.PrintDVector(dvect)


lsci.DVectorHasValue.argtypes = [ctypes.POINTER(DVECTOR), ctypes.c_double]
lsci.DVectorHasValue.restype = ctypes.c_int


def dvector_has_value(dvect, val):
    """
    DVectorHasValue: Check if a libscientific double vector contains an exact
                     value "val" and return 0 or 1 respectivelly for have or not have.
    """
    return lsci.DVectorHasValue(dvect, val)


lsci.DVectorSet.argtypes = [ctypes.POINTER(DVECTOR), ctypes.c_double]
lsci.DVectorSet.restype = None


def dvector_set(dvect, val):
    """
    DVectorSet: Set all values of a libscientific double vector to "val"
    """
    lsci.DVectorSet(dvect, val)


lsci.setDVectorValue.argtypes = [ctypes.POINTER(DVECTOR),
                                 ctypes.c_size_t,
                                 ctypes.c_double]
lsci.setDVectorValue.restype = None


def set_dvector_value(dvect, indx, val):
    """
    setDVectorValue: Set/modify a value in the row of a libscientific
                     double vector
    """
    lsci.setDVectorValue(dvect, indx, val)


lsci.getDVectorValue.argtypes = [ctypes.POINTER(DVECTOR),
                                 ctypes.c_size_t]
lsci.getDVectorValue.restype = ctypes.c_double


def get_dvector_value(dvect, indx):
    """
    getDVectorValue: Get a value in the row of a libscientific
                     double vector
    """
    return lsci.getDVectorValue(dvect, indx)


def dvector_tolist(dvect):
    """
    Convert a libscientific double vector into a python list
    """
    d_lst = []
    try:
        for i in range(dvect[0].size):
            d_lst.append(dvect[0].data[i])
    except TypeError:
        for i in range(dvect.size):
            d_lst.append(dvect.data[i])
    return d_lst


lsci.DVectorAppend.argtypes = [ctypes.POINTER(DVECTOR),
                               ctypes.c_double]
lsci.DVectorAppend.restype = None

def dvector_append(dvect, val):
    """
    Append a value to a double vector d
    """
    return lsci.DVectorAppend(dvect, val)


lsci.DVectorRemoveAt.argtypes = [ctypes.POINTER(DVECTOR),
                                 ctypes.c_size_t]
lsci.DVectorRemoveAt.restype = None

def dvector_remove_at(dvect, indx):
    """
    Remove a value from a double vector d at index indx
    """
    return lsci.DVectorRemoveAt(dvect, indx)

lsci.DVectorCopy.argtypes = [ctypes.POINTER(DVECTOR),
                             ctypes.POINTER(DVECTOR)]
lsci.DVectorCopy.restype = None

def dvector_copy(dvect_src):
    """
    Create a copy of dvector d to a
    """
    dvect_dst = init_dvector()
    lsci.DVectorCopy(dvect_src, dvect_dst)
    return dvect_dst


class DVector():
    """
    Translate a list  into a libscientific double vector
    """
    def __init__(self, dvect_):
        self.dvect = new_dvector(dvect_)

    def __del__(self):
        del_dvector(self.dvect)
        del self.dvect

    def __getitem__(self, key):
        return self.data_ptr()[key]

    def __setitem__(self, key, value):
        set_dvector_value(self.dvect, key, value)

    def size(self):
        """
        return the size of the dvector
        """
        return self.dvect[0].size

    def data_ptr(self):
        """
        return the pointer to data
        """
        return self.dvect[0].data

    def append(self, value):
        """
        Append a value to the dvector
        """
        return dvector_append(self.dvect, value)

    def extend(self, v_lst):
        """
        Extend the dvector by adding a list
        """
        for val in v_lst:
            dvector_append(self.dvect, val)

    def tolist(self):
        """
        Convert the dvector to a list
        """
        return dvector_tolist(self.dvect)

    def fromlist(self, vlst_):
        """
        Convert a list to a dvector
        """
        del_dvector(self.dvect)
        del self.dvect
        self.dvect = new_dvector(vlst_)

    def debug(self):
        """
        Debug the dvector
        """
        print_dvector(self.dvect)


if __name__ in "__main__":
    from random import random
    a = [random() for j in range(10)]
    d = DVector(a)
    d.debug()
    print("get value")
    print(d[1])
    print("set value")
    d[1] = -2
    dlst = d.tolist()
    print("print the list converted from the double vector")
    for item in dlst:
        print(item)

    print("Add at the end the value -123")
    dvector_append(d.dvect, -123)
    d.debug()

    print("Append in a different way -123 at the end")
    d.append(-123)
    d.debug()

    print("remove at index 1, then value -2")
    dvector_remove_at(d.dvect, 1)
    d.debug()

    print("Extend d with b")
    print("b:")
    b = [random() for j in range(4)]
    print(b)
    d.extend(b)
    print("d extended:")
    d.debug()

    print("Create a copy of d.d in q")
    q = dvector_copy(d.dvect)
    print_dvector(q)
    del_dvector(q)

    print("Check if the double vector d have the value -123")
    print(dvector_has_value(d.dvect, -13.0000))



"""
vector libscientific python binding

Copyright (C) <2022>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
from libscientific import vector as vect
from libscientific.loadlibrary import load_libscientific_library

lsci = load_libscientific_library()


class DVECTLIST(ctypes.Structure):
    """
    DVECTLIST data structure
    """
    _fields_ = [
        ("dvector", ctypes.POINTER(ctypes.POINTER(vect.DVECTOR))),
        ("size", ctypes.c_size_t)]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__


lsci.initDVectorList.argtypes = [ctypes.POINTER(ctypes.POINTER(DVECTLIST))]
lsci.initDVectorList.restype = None


def init_dvector_list():
    """
    init_dvector_list: Allocate in memory an empty libscientific double vector list
    """
    dvl = ctypes.POINTER((DVECTLIST))()
    lsci.initDVectorList(ctypes.pointer(dvl))
    return dvl


def new_dvector_list(dvl_):
    """
    new_dvector_list: Allocate in memory a libscientific double vector list
    from a list of lists
    """
    dvl = init_dvector_list()
    for v_lst in dvl_:
        dvector_list_append(dvl, v_lst)
    return dvl


lsci.DelDVectorList.argtypes = [ctypes.POINTER(ctypes.POINTER(DVECTLIST))]
lsci.DelDVectorList.restype = None


def del_dvector_list(dvl):
    """
    del_dvector_list: Delete an allocated libscientific double vector
    """
    lsci.DelDVectorList(ctypes.pointer(dvl))


lsci.DVectorListAppend.argtypes = [ctypes.POINTER(DVECTLIST),
                                   ctypes.POINTER(vect.DVECTOR)]
lsci.DVectorListAppend.restype = None

def dvector_list_append(dvl, dvect):
    """
    dvector_list_append: Append a double vector to a DvectorList
    """
    if isinstance(dvect, list):
        dvect_ = vect.new_dvector(dvect)
        lsci.DVectorListAppend(dvl, dvect_)
        vect.del_dvector(dvect_)
        del dvect_
    else:
        lsci.DVectorListAppend(dvl, dvect)
    return 0


def dvector_list_tolist(dvl):
    """
    convert a divector list into a list of lists
    """
    lsts = []
    for i in range(dvl[0].size):
        lsts.append(vect.dvector_tolist(dvl[0].dvector[i].contents))
    return lsts


class DVectorList():
    """
    Translate a list  into a libscientific double vector
    """
    def __init__(self, dvl_):
        self.dvl = new_dvector_list(dvl_)

    def __del__(self):
        del_dvector_list(self.dvl)
        del self.dvl

    def __getitem__(self, key):
        return self.data_ptr()[key].contents

    def __setitem__(self, key, v_lst):
        for i, val in enumerate(v_lst):
            if i < self.data_ptr()[key].contents.size:
                self.data_ptr()[key].contents.data[i] = val
            else:
                break
        return 0

    def size(self):
        """
        return the size of the divector
        """
        return self.dvl[0].size

    def data_ptr(self):
        """
        return the pointer to data
        """
        return self.dvl[0].dvector

    def append(self, v_lst):
        """
        Append a value to the dvector
        """
        if isinstance(lst, vect.DVECTOR):
            dvector_list_append(self.dvl, v_lst)
        else:
            dv_lst = vect.new_dvector(lst)
            dvector_list_append(self.dvl, dv_lst)
            vect.del_dvector(dv_lst)
            del dv_lst
        return 0

    def tolist(self):
        """
        Convert the dvector list to a list of list
        """
        return dvector_list_tolist(self.dvl)

    def fromlist(self, v_lists):
        """
        Convert a list of list to a dvector list
        """
        for v_lst in v_lists:
            dvector_list_append(self.dvl, v_lst)
        return 0

    def debug(self):
        """
        Debug the double list vector
        """
        for i in range(self.dvl[0].size):
            vect.print_dvector(self.dvl[0].dvector[i].contents)


if __name__ in "__main__":
    from random import random
    a = [[random() for j in range(10)] for i in range(3)]
    d = DVectorList(a)
    d.debug()
    print("get value")
    vect.print_dvector(d[1])
    print("set list")
    d[1] = [8, 8, 8, 8, 8, 8, 8, 8, 8, 8]
    dlst = d.tolist()
    print("print the list converted")
    for item in dlst:
        print(item)

    print("Append a list - high level")
    lst = [9, 8, 7, 6]
    d.append(lst)
    print("Reappend ppend a list - low level")
    dvector_list_append(d.dvl, lst)
    d.debug()



"""
info libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
from libscientific.loadlibrary import load_libscientific_library

lsci = load_libscientific_library()

lsci.GetScientificVersion.argtypes = (None)
lsci.GetScientificVersion.restype = ctypes.c_char_p


class Scientific():
    """
    Libscientific info
    """
    def __init__(self):
        self.__version__ = str(lsci.GetScientificVersion().decode('UTF-8'))

    def print_version(self):
        """
        Print to video libscientific version
        """
        print(self.__version__)

    def version(self):
        """
        Get libscientific version as string
        """
        return self.__version__

if __name__ == '__main__':
    libscientific = Scientific()
    print(libscientific.__version__)



"""
loadlibrary libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from ctypes import cdll
from ctypes.util import find_library

def load_libscientific_library():
    """
    Load the libscientific library
    """
    lsci = None
    library = find_library("scientific")
    if library is None:
        message = "Please install libscientific. "
        message += "Go to https://github.com/gmrandazzo/libscientific"
        message += "or if it is installed please check/specify the location in LD_LIBRARY_PATH"
        print(message)
    else:
        lsci = cdll.LoadLibrary(library)
    return lsci


if __name__ == '__main__':
    a = load_libscientific_library()



"""
cpca libscientific python binding

Copyright (C) <2022>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
import libscientific.vectlist as vlst
import libscientific.matrix as mx
import libscientific.tensor as tns
import libscientific.vector as vect
from libscientific.loadlibrary import load_libscientific_library

lsci = load_libscientific_library()


class CPCAMODEL(ctypes.Structure):
    """
    CPCA model data structure
    """
    _fields_ = [
        ("block_scores", ctypes.POINTER(tns.TENSOR)),
        ("block_loadings", ctypes.POINTER(tns.TENSOR)),
        ("super_scores", ctypes.POINTER(mx.MATRIX)),
        ("super_weights", ctypes.POINTER(mx.MATRIX)),
        ("scaling_factor", ctypes.POINTER(vect.DVECTOR)),
        ("total_expvar", ctypes.POINTER(vect.DVECTOR)),
        ("block_expvar", ctypes.POINTER(vlst.DVECTLIST)),
        ("colaverage", ctypes.POINTER(vlst.DVECTLIST)),
        ("colscaling", ctypes.POINTER(vlst.DVECTLIST))]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__



lsci.NewCPCAModel.argtypes = [ctypes.POINTER(ctypes.POINTER(CPCAMODEL))]
lsci.NewCPCAModel.restype = None


def new_cpca_model():
    """
    new_cpca_model: Allocate in memory an empty libscientific CPCA model
    """
    mpca = ctypes.POINTER(CPCAMODEL)()
    lsci.NewCPCAModel(ctypes.pointer(mpca))
    return mpca


lsci.DelCPCAModel.argtypes = [ctypes.POINTER(ctypes.POINTER(CPCAMODEL))]
lsci.DelCPCAModel.restype = None


def del_cpca_model(mpca):
    """
    del_cpca_model: Delete an allocated libscientific CPCA model
    """
    lsci.DelCPCAModel(ctypes.pointer(mpca))


lsci.CPCA.argtypes = [ctypes.POINTER(tns.TENSOR),
                     ctypes.c_size_t,
                     ctypes.c_size_t,
                     ctypes.POINTER(CPCAMODEL),
                     ctypes.POINTER(ctypes.c_int)]
lsci.CPCA.restype = None


def cpca_algorithm(t_input, scaling, npc, mpca):
    """
    CPCA: Calculate the CPCA model for a matrix m using the NIPALS algorithm
    """
    ssignal = ctypes.c_int(0)
    lsci.CPCA(t_input, scaling, npc, mpca, ssignal)


lsci.CPCAScorePredictor.argtypes = [ctypes.POINTER(tns.TENSOR),
                                   ctypes.POINTER(CPCAMODEL),
                                   ctypes.c_size_t,
                                   ctypes.POINTER(mx.MATRIX),
                                   ctypes.POINTER(tns.TENSOR)]
lsci.CPCAScorePredictor.restype = None


def cpca_score_predictor(t_input, model, npc, p_super_scores, p_block_scores):
    """
    cpca_score_predictor: Predict scores for a matrix m in the computed CPCA modes
    """
    lsci.CPCAScorePredictor(t_input,
                            model,
                            npc,
                            p_super_scores,
                            p_block_scores)


lsci.PrintCPCA.argtypes = [ctypes.POINTER(CPCAMODEL)]
lsci.PrintCPCA.restype = None

def print_cpca(mpca):
    """
    PrintCPCA: Print to video the CPCA Model
    """
    lsci.PrintCPCA(mpca)


class CPCA():
    """
    CPCA model class
    """
    def __init__(self, scaling, npc):
        self.model = new_cpca_model()
        self.scaling = scaling
        self.npc = npc

    def __del__(self):
        if self.model is not None:
            del_cpca_model(self.model)
            del self.model
        self.model = None

    def fit(self, t_input):
        """
        fit a cpca model using an input tensor
        """
        if "Tensor" in str(type(t_input)):
            cpca_algorithm(t_input, self.scaling, self.npc, self.model)
        else:
            t_input_ = tns.new_tensor(t_input)
            cpca_algorithm(t_input_, self.scaling, self.npc, self.model)
            tns.del_tensor(t_input_)
            del t_input_

    def get_super_scores(self):
        """
        get the cpca super scores
        """
        return mx.matrix_to_list(self.model[0].super_scores)

    def get_super_weights(self):
        """
        get the cpca super weights
        """
        return mx.matrix_to_list(self.model[0].super_weights)

    def get_block_scores(self):
        """
        get the cpca block scores
        """
        return tns.tensor_tolist(self.model[0].block_scores)

    def get_block_loadings(self):
        """
        get the cpca block loadings
        """
        return tns.tensor_tolist(self.model[0].block_loadings)

    def get_block_expvar(self):
        """
        get the cpca block variance explained
        """
        return vlst.dvector_list_tolist(self.model[0].block_expvar)

    def get_total_exp_variance(self):
        """
        get the cpca total variance explained
        """
        return vect.dvector_tolist(self.model[0].total_expvar)

    def predict(self, t_input):
        """
        get the cpca block variance explained
        """
        t_input_ = tns.new_tensor(t_input)
        p_super_scores_ = mx.init_matrix()
        p_block_scores_ = tns.init_tensor()

        cpca_score_predictor(t_input_,
                            self.model,
                            self.npc,
                            p_super_scores_,
                            p_block_scores_)
        p_super_scores = mx.matrix_to_list(p_super_scores_)
        p_block_scores = tns.tensor_tolist(p_block_scores_)
        mx.del_matrix(p_super_scores_)
        del p_super_scores_

        tns.del_tensor(p_block_scores_)
        del p_block_scores_

        tns.del_tensor(t_input_)
        del t_input_
        return p_super_scores, p_block_scores


if __name__ == '__main__':
    def mx_to_video(m_input, decimals=5):
        """
        print a matrix to video
        """
        for row in m_input:
            print("\t".join([str(round(x, decimals)) for x in row]))

    def t_to_video(t_input):
        """
        print a tensor to video
        """
        i = 1
        for mx_input in t_input:
            print(f"Block: {i}")
            mx_to_video(mx_input, 3)
            i+=1

    import random
    random.seed(123456)
    a = [[[random.random() for j in range(4)] for i in range(10)] for k in range(4)]
    t = tns.new_tensor(a)
    print("Original Tensor")
    t_to_video(a)
    print("Computing CPCA ...")
    cpca = CPCA(1, 2)
    cpca.fit(a)
    print("Showing the CPCA super scores")
    sscores = cpca.get_super_scores()
    mx_to_video(sscores, 3)
    print("Showing the CPCA super weights")
    sweights = cpca.get_super_weights()
    mx_to_video(sweights, 3)
    print("Showing the CPCA block scores")
    block_scores = cpca.get_block_scores()
    t_to_video(block_scores)
    print("Showing the CPCA block loadings")
    block_loadings = cpca.get_block_loadings()
    t_to_video(block_loadings)

    print(cpca.get_total_exp_variance())
    p_ss, p_bs = cpca.predict(a)
    mx_to_video(p_ss, 3)



"""
misc libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
from libscientific.loadlibrary import load_libscientific_library

lsci = load_libscientific_library()


lsci.missing_value.argtypes = None
lsci.missing_value.restype = ctypes.c_double


def missing_value():
    """
    missing_value: return the standard missing value in
    libscientific. T^his variable is useful to
    compute PCA or PLS with missing values.
    """
    return lsci.missing_value()



"""
pls libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
import libscientific.matrix as mx
import libscientific.vector as vect
import libscientific.tensor as t
from libscientific.loadlibrary import load_libscientific_library

lsci = load_libscientific_library()


class PLSMODEL(ctypes.Structure):
    """
    PLSMODEL data structure
    """
    _fields_ = [
        ("xscores", ctypes.POINTER(mx.MATRIX)),
        ("xloadings", ctypes.POINTER(mx.MATRIX)),
        ("xweights", ctypes.POINTER(mx.MATRIX)),
        ("yscores", ctypes.POINTER(mx.MATRIX)),
        ("yloadings", ctypes.POINTER(mx.MATRIX)),
        ("cweights", ctypes.POINTER(mx.MATRIX)),
        ("b", ctypes.POINTER(vect.DVECTOR)),
        ("xvarexp", ctypes.POINTER(vect.DVECTOR)),
        ("xcolaverage", ctypes.POINTER(vect.DVECTOR)),
        ("xcolscaling", ctypes.POINTER(vect.DVECTOR)),
        ("ycolaverage", ctypes.POINTER(vect.DVECTOR)),
        ("ycolscaling", ctypes.POINTER(vect.DVECTOR)),
        ("recalculated_y", ctypes.POINTER(mx.MATRIX)),
        ("recalc_residuals", ctypes.POINTER(mx.MATRIX)),
        ("predicted_y", ctypes.POINTER(mx.MATRIX)),
        ("pred_residuals", ctypes.POINTER(mx.MATRIX)),
        ("recalc_residuals", ctypes.POINTER(mx.MATRIX)),
        ("r2y_recalculated", ctypes.POINTER(mx.MATRIX)),
        ("r2y_validation", ctypes.POINTER(mx.MATRIX)),
        ("q2y", ctypes.POINTER(mx.MATRIX)),
        ("sdep", ctypes.POINTER(mx.MATRIX)),
        ("sdec", ctypes.POINTER(mx.MATRIX)),
        ("bias", ctypes.POINTER(mx.MATRIX)),
        ("roc_recalculated", ctypes.POINTER(t.TENSOR)),
        ("roc_validation", ctypes.POINTER(t.TENSOR)),
        ("roc_auc_recalculated", ctypes.POINTER(mx.MATRIX)),
        ("roc_auc_validation", ctypes.POINTER(mx.MATRIX)),
        ("roc_auc_validation", ctypes.POINTER(mx.MATRIX)),
        ("precision_recall_recalculated", ctypes.POINTER(t.TENSOR)),
        ("precision_recall_validation", ctypes.POINTER(t.TENSOR)),
        ("precision_recall_ap_recalculated", ctypes.POINTER(mx.MATRIX)),
        ("precision_recall_ap_validation", ctypes.POINTER(mx.MATRIX)),
        ("yscrambling", ctypes.POINTER(mx.MATRIX))]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__


lsci.NewPLSModel.argtypes = [ctypes.POINTER(ctypes.POINTER(PLSMODEL))]
lsci.NewPLSModel.restype = None


def new_pls_model():
    """
    new_pls_model: Allocate in memory an empty libscientific PLS model
    """
    mpls = ctypes.POINTER(PLSMODEL)()
    lsci.NewPLSModel(ctypes.pointer(mpls))
    return mpls


lsci.DelPLSModel.argtypes = [ctypes.POINTER(ctypes.POINTER(PLSMODEL))]
lsci.DelPLSModel.restype = None


def del_pls_model(mpls):
    """
    del_pls_model: Delete an allocated libscientific PLS model
    """
    lsci.DelPLSModel(ctypes.pointer(mpls))


lsci.PLS.argtypes = [ctypes.POINTER(mx.MATRIX),
                     ctypes.POINTER(mx.MATRIX),
                     ctypes.c_size_t,
                     ctypes.c_size_t,
                     ctypes.c_size_t,
                     ctypes.POINTER(PLSMODEL),
                     ctypes.POINTER(ctypes.c_int)]
lsci.PLS.restype = None


def pls_algorithm(x_input, y_input, nlv, x_scaling, y_scaling, mpls):
    """
    PLS: Calculate the PLS model using a matrix x and a matrix y
         according to the NIPALS algorithm

    PLS(matrix *mx,
        matrix *my,
        size_t nlv,
        size_t xautoscaling,
        size_t yautoscaling,
        PLSMODEL* model,
        ssignal *s)
    """
    ssignal = ctypes.c_int(0)
    lsci.PLS(x_input,
             y_input,
             nlv,
             x_scaling,
             y_scaling,
             mpls,
             ctypes.pointer(ssignal))


lsci.PLSBetasCoeff.argtypes = [ctypes.POINTER(PLSMODEL),
                               ctypes.c_size_t,
                               ctypes.POINTER(vect.DVECTOR)]
lsci.PLSBetasCoeff.restype = None


def pls_beta_coefficients(mpls, nlv, bcoeff):
    """
    PLSBetasCoeff calculation
    """
    lsci.PLSBetasCoeff(mpls, nlv, ctypes.pointer(bcoeff))



lsci.PLSScorePredictor.argtypes = [ctypes.POINTER(mx.MATRIX),
                                   ctypes.POINTER(PLSMODEL),
                                   ctypes.c_size_t,
                                   ctypes.POINTER(mx.MATRIX)]
lsci.PLSScorePredictor.restype = None


def pls_score_predictor(x_input, mpls, nlv, p_scores):
    """
    PLSScorePredictor: Predict scores for a matrix m in the computed PLS modes
    """
    lsci.PLSScorePredictor(x_input,
                           mpls,
                           nlv,
                           p_scores)


lsci.PLSYPredictor.argtypes = [ctypes.POINTER(mx.MATRIX),
                               ctypes.POINTER(PLSMODEL),
                               ctypes.c_size_t,
                               ctypes.POINTER(mx.MATRIX)]
lsci.PLSYPredictor.restype = None


def pls_y_predictor(x_scores, mpls, nlv, predicted_y):
    """
    PLSYPredictor: Predict the Y according the predicted scores and the calculated pls model.
    """
    lsci.PLSYPredictor(x_scores,
                       mpls,
                       nlv,
                       ctypes.pointer(ctypes.pointer(predicted_y)))


# void PLSYPredictorAllLV(matrix *mx, PLSMODEL *model, matrix **tscores, matrix **y);

lsci.PLSYPredictorAllLV.argtypes = [ctypes.POINTER(mx.MATRIX),
                                    ctypes.POINTER(PLSMODEL),
                                    ctypes.POINTER(mx.MATRIX),
                                    ctypes.POINTER(mx.MATRIX)]
lsci.PLSYPredictorAllLV.restype = None


def pls_y_predictor_all_lv(x_input, mpls, predicted_scores, predicted_y):
    """
    PLSYPredictorAllLV: Predict the Y according the original
                        feature matrix and the calculated pls model.
                        This function is NOT dependent on PLSScorePredictor.
    """
    lsci.PLSYPredictorAllLV(x_input,
                            mpls,
                            predicted_scores,
                            predicted_y)



lsci.PrintPLSModel.argtypes = [ctypes.POINTER(PLSMODEL)]
lsci.PrintPLSModel.restype = None


def print_pls(mpls):
    """
    PrintPLS: Print to video the PLS Model
    """
    lsci.PrintPLSModel(mpls)


class PLS():
    """
    Partial least squares
        Arguments:
            nlv (int) : number of latent variable
            xscaling (int) : scaling type for x matrix
            yscaling (int) : scaling type for y matrix
            N.B.: xscaling and yscaling can be:
                0 -> No scaling
                1 -> Standard deviation scaling
                2 -> Root mean squared scaling
                3 -> Pareto scaling
                4 -> Range scaling
                5 -> Level scaling
    """
    def __init__(self, nlv, xscaling=1, yscaling=0):
        self.mpls = new_pls_model()
        self.nlv = nlv
        self.xscaling = xscaling
        self.yscaling = yscaling

    def __del__(self):
        if self.mpls is not None:
            del_pls_model(self.mpls)
            del self.mpls
        self.mpls = None

    def fit(self, x_input, y_input):
        """
        Fit a pls model giving x matrix and y matrix
        """
        x_input_ = None
        xalloc = False
        if "Matrix" not in str(type(x_input)):
            x_input_ = mx.new_matrix(x_input)
            xalloc = True
        else:
            x_input_ = x_input

        y_input_ = None
        yalloc = False
        if "Matrix" not in str(type(y_input)):
            y_input_ = mx.new_matrix(y_input)
            yalloc = True
        else:
            y_input_ = y_input

        pls_algorithm(x_input_,
                      y_input_,
                      self.nlv,
                      self.xscaling,
                      self.yscaling,
                      self.mpls)

        if xalloc is True:
            mx.del_matrix(x_input_)
            del x_input_

        if yalloc is True:
            mx.del_matrix(y_input_)
            del y_input_

    def get_tscores(self):
        """
        Get the T-Scores
        """
        return mx.matrix_to_list(self.mpls[0].xscores)

    def get_uscores(self):
        """
        Get the U-Scores
        """
        return mx.matrix_to_list(self.mpls[0].yscores)

    def get_ploadings(self):
        """
        Get the P-Loadings
        """
        return mx.matrix_to_list(self.mpls[0].xloadings)

    def get_qloadings(self):
        """
        Get the Q-Loadings
        """
        return mx.matrix_to_list(self.mpls[0].yloadings)

    def get_weights(self):
        """
        Get the W-weigths
        """
        return mx.matrix_to_list(self.mpls[0].xweights)

    def get_exp_variance(self):
        """
        Get the explained variance
        """
        return vect.dvector_tolist(self.mpls[0].xvarexp)

    def predict(self, x_input, nlv_=None):
        """
        Predict the y giving an x_ matrix
        """
        x_input_ = mx.new_matrix(x_input)
        p_scores_ = mx.init_matrix()
        p_y_ = mx.init_matrix()


        pls_y_predictor_all_lv(x_input_, self.mpls, p_scores_, p_y_)
        p_scores = mx.matrix_to_list(p_scores_)

        p_y = None
        if nlv_ is None:
            p_y = mx.matrix_to_list(p_y_)
        else:
            p_y = [[row[nlv_-1]] for row in mx.matrix_to_list(p_y_)]
        mx.del_matrix(x_input_)
        del x_input_
        mx.del_matrix(p_scores_)
        del p_scores_
        mx.del_matrix(p_y_)
        del p_y_
        return p_y, p_scores


if __name__ == '__main__':
    def mx_to_video(m_input, decimals=5):
        """
        print a matrix to video
        """
        for row in m_input:
            print("\t".join([str(round(x, decimals)) for x in row]))
    import random
    random.seed(123456)
    x = [[random.random() for j in range(4)] for i in range(10)]
    y = [[random.random() for j in range(1)] for i in range(10)]
    xp = [[random.random() for j in range(4)] for i in range(10)]

    print("Original Matrix")
    print("X")
    mx_to_video(x)
    print("Y")
    mx_to_video(y)
    print("XP")
    mx_to_video(xp)
    print("Computing PLS ...")
    model = PLS(nlv=2, xscaling=1, yscaling=0)
    model.fit(x, y)
    print("Showing the PLS T scores")
    tscores = model.get_tscores()
    mx_to_video(tscores, 3)

    print("Showing the PLS U scores")
    uscores = model.get_uscores()
    mx_to_video(uscores, 3)

    print("Showing the PLS P loadings")
    ploadings = model.get_ploadings()
    mx_to_video(ploadings, 3)

    print("Showing the X Variance")
    print(model.get_exp_variance())


    print("Predict XP")
    py, pscores = model.predict(xp)
    print("Predicted Y for all LVs")
    mx_to_video(py, 3)
    print("Predicted Scores")
    mx_to_video(pscores, 3)



from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='libscientific',
      version='1.5.0',
      description='Libscientific python foreign function interface',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/gmrandazzo/libscientific',
      author='Giuseppe Marco Randazzo',
      author_email='gmrandazzo@gmail.com',
      license='GPLv3',
      packages=['libscientific'],
      python_requires=">=3.0",
      zip_safe=False)



"""
tensor libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
from libscientific.loadlibrary import load_libscientific_library
from libscientific import misc
from libscientific import matrix as mx
from libscientific import vector as vect

lsci = load_libscientific_library()

class TENSOR(ctypes.Structure):
    """
    tensor data structure
    """
    _fields_ = [
        ("m", ctypes.POINTER(ctypes.POINTER(mx.MATRIX))),
        ("order", ctypes.c_size_t)]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__

lsci.initTensor.argtypes = [ctypes.POINTER(ctypes.POINTER(TENSOR))]
lsci.initTensor.restype = None

def init_tensor():
    """
    initTensor: Allocate in memory an empty libscientific tensor
    """
    tns = ctypes.POINTER(TENSOR)()
    lsci.initTensor(ctypes.pointer(tns))
    return tns


lsci.NewTensorMatrix.argtypes = [ctypes.POINTER(TENSOR),
                                 ctypes.c_size_t,
                                 ctypes.c_size_t,
                                 ctypes.c_size_t]
lsci.NewTensorMatrix.restype = None

def new_tensor_matrix(tns, k_indx, num_row, num_col):
    """
    Create a new matrix on tensor tns at index k_indx
    """
    lsci.NewTensorMatrix(tns, k_indx, num_row, num_col)


lsci.AddTensorMatrix.argtypes = [ctypes.POINTER(TENSOR),
                                 ctypes.c_size_t,
                                 ctypes.c_size_t]
lsci.AddTensorMatrix.restype = None

def add_tensor_matrix(tns, num_row, num_col):
    """
    Add a new matrix to the tensor tns
    """
    lsci.AddTensorMatrix(tns, num_row, num_col)


lsci.NewTensor.argtypes = [ctypes.POINTER(ctypes.POINTER(TENSOR)),
                           ctypes.c_size_t]
lsci.NewTensor.restype = None

def new_tensor(tns_input_):
    """
    NewTensor: Allocate in memory a libscientific tensor
    from a list of lists of lists
    """
    tns_input = None
    if "numpy" in str(type(tns_input_)):
        tns_input = tns_input_.tolist()
    else:
        tns_input = tns_input_
    order = len(tns_input)
    tns = ctypes.POINTER(TENSOR)()
    lsci.NewTensor(ctypes.pointer(tns), order)
    for k in range(order):
        nrows = len(tns_input[k])
        ncols = len(tns_input[k][0])
        lsci.NewTensorMatrix(tns, k, nrows, ncols)
        for i in range(nrows):
            for j in range(ncols):
                val = None
                try:
                    val = float(tns_input[k][i][j])
                except ValueError:
                    val = None

                if val is None:
                    tns.contents.m[k].contents.data[i][j] = misc.missing_value()
                else:
                    tns.contents.m[k].contents.data[i][j] = val
    return tns


lsci.DelTensor.argtypes = [ctypes.POINTER(ctypes.POINTER(TENSOR))]
lsci.DelTensor.restype = None

def del_tensor(tns):
    """
    Delete the moemory allocated tensor
    """
    lsci.DelTensor(ctypes.pointer(tns))


lsci.PrintTensor.argtypes = [ctypes.POINTER(TENSOR)]
lsci.PrintTensor.restype = None

def print_tensor(tns):
    """
    Debug the tensor content
    """
    lsci.PrintTensor(tns)


lsci.setTensorValue.argtypes = [ctypes.POINTER(TENSOR)]
lsci.setTensorValue.restype = None

def set_tensor_value(tns, k_indx, i_indx, j_indx, val):
    """
    Set the tensor value at indexes k_indx, i_indx, j_indx
    """
    lsci.setTensorValue(tns, k_indx, i_indx, j_indx, val)


lsci.TensorAppendColumn.argtypes = [ctypes.POINTER(TENSOR),
                                    ctypes.c_size_t,
                                    ctypes.POINTER(vect.DVECTOR)]
lsci.TensorAppendColumn.restype = None

def tensor_append_column(tns, k_indx, dvect):
    """
    Append a colum to the matrix at index k_indx inside the tensor tns
    """
    lsci.TensorAppendColumn(tns, k_indx, dvect)


lsci.TensorAppendRow.argtypes = [ctypes.POINTER(TENSOR),
                                 ctypes.c_size_t,
                                 ctypes.POINTER(vect.DVECTOR)]
lsci.TensorAppendRow.restype = None

def tensor_append_row(tns, k_indx, dvect):
    """
    Append a row to the matrix at index k_indx inside the tensor tns
    """
    lsci.TensorAppendRow(tns, k_indx, dvect)


lsci.TensorSet.argtypes = [ctypes.POINTER(TENSOR),
                           ctypes.c_double]
lsci.TensorSet.restype = None

def tensor_set(tns, val):
    """
    Set all the tensor values to a given value val
    """
    lsci.TensorSet(tns, val)


lsci.TensorCopy.argtypes = [ctypes.POINTER(TENSOR),
                           ctypes.POINTER(ctypes.POINTER(TENSOR))]
lsci.TensorCopy.restype = None

def tensor_copy(src_tns, dst_tns):
    """
    Copy a source tensor to a destination tensor
    """
    lsci.TensorCopy(src_tns, ctypes.pointer(dst_tns))


def tensor_tolist(tns):
    """
    Convert a tensor to list
    """
    tns_lst = []
    for k in range(tns.contents.order):
        tns_lst.append(mx.matrix_to_list(tns.contents.m[k]))
    return tns_lst


class Tensor():
    """
    Translate a list of list of list into a libscientific tensor
    """
    def __init__(self, tns_):
        self.tns = new_tensor(tns_)

    def __del__(self):
        del_tensor(self.tns)
        del self.tns
        self.tns = None

    def __getitem__(self, keys):
        k, i, j = keys
        return self.data_ptr().m[k].contents.data[i][j]

    def __setitem__(self, keys, value):
        k, i, j = keys
        self.data_ptr().m[k].contents.data[i][j] = value

    def order(self):
        """
        Return the order of the tensor aka the number of matrix
        constituting the tensor
        """
        return self.data_ptr().order

    def nrow(self, k):
        """
        Return the number of rows of the matrix k of the tensor
        """
        return self.data_ptr().m[k].contents.row

    def ncol(self, k):
        """
        Return the number of column of the matrix k of the tensor
        """
        return self.data_ptr().m[k].contents.col

    def data_ptr(self):
        """
        Return the tensor data pointer
        """
        return self.tns.contents

    def tolist(self):
        """
        Return the tensor as list
        """
        return tensor_tolist(self.tns)

    def fromlist(self, tns_):
        """
        Get a tensor fro list
        """
        del_tensor(self.tns)
        del self.tns
        self.tns = new_tensor(tns_)

    def fromnumpy(self, npt):
        """
        Get a tensor from numpy
        """
        tns_ = npt.tolist()
        del_tensor(self.tns)
        del self.tns
        self.tns = new_tensor(tns_)

    def debug(self):
        """
        Debug to video the tensor
        """
        print_tensor(self.tns)



if __name__ in "__main__":
    print("Tensort Test")
    a = [[[1,2,3],[1,2,3]],[[4,5,6],[4,5,6]]]
    t = new_tensor(a)
    print_tensor(t)
    lst = tensor_tolist(t)
    print(lst)
    add_tensor_matrix(t, 5, 6)
    print_tensor(t)
    b = init_tensor()
    tensor_copy(t, b)
    print_tensor(b)
    del_tensor(t)
    del_tensor(b)

    t = init_tensor()
    add_tensor_matrix(t, 3, 2)
    print_tensor(t)
    del_tensor(t)

    t = Tensor(a)
    print(type(t))
    print(t.order())
    print(t.nrow(0))
    print(t.ncol(0))
    print(t.tolist())
    a[0][0][0] = -9876
    t.fromlist(a)
    t.debug()

    import numpy as np
    a = np.array(a)
    a[0][1][1] = 765
    t.fromnumpy(a)
    t.debug()



"""
matrix libscientific python binding

Copyright (C) <2019>  Giuseppe Marco Randazzo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import ctypes
from libscientific.loadlibrary import load_libscientific_library
from libscientific import misc
from libscientific import vector

lsci = load_libscientific_library()

class MATRIX(ctypes.Structure):
    """
    matrix data structure
    """
    _fields_ = [
        ("data",    ctypes.POINTER(ctypes.POINTER(ctypes.c_double))),
        ("row",     ctypes.c_size_t),
        ("col",     ctypes.c_size_t)]

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__


lsci.initMatrix.argtypes = [ctypes.POINTER(ctypes.POINTER(MATRIX))]
lsci.initMatrix.restype = None


def init_matrix():
    """
    init_matrix: Allocate in memory an empty libscientific matrix
    """
    mtx = ctypes.POINTER(MATRIX)()
    # m = matrix()
    lsci.initMatrix(ctypes.pointer(mtx))
    return mtx


lsci.NewMatrix.argtypes = [ctypes.POINTER(ctypes.POINTER(MATRIX)),
                           ctypes.c_size_t,
                           ctypes.c_size_t]
lsci.NewMatrix.restype = None


def new_matrix(mtx_input_):
    """
    new_matrix: Allocate in memory a libscientific matrix from a list of lists
    """
    mtx_input = None
    if "numpy" in str(type(mtx_input_)):
        mtx_input = mtx_input_.tolist()
    else:
        mtx_input = mtx_input_
    nrows = None
    ncols = None
    try:
        nrows = len(mtx_input)
        try:
            ncols = len(mtx_input[0])
        except IndexError:
            ncols = 0
    except IndexError:
        nrows = 0

    mtx = ctypes.POINTER(MATRIX)()
    lsci.NewMatrix(ctypes.pointer(mtx),
                   nrows,
                   ncols)

    for i in range(nrows):
        for j in range(ncols):
            val = None
            try:
                val = float(mtx_input[i][j])
            except ValueError:
                val = None

            if val is None:
                set_missing_matrix_value(mtx, i, j)
            else:
                lsci.setMatrixValue(mtx, i, j, val)
    return mtx


lsci.ResizeMatrix.argtypes = [ctypes.POINTER(MATRIX),
                              ctypes.c_size_t,
                              ctypes.c_size_t]
lsci.ResizeMatrix.restype = None


def resize_matrix(mtx, nrows, ncols):
    """
    resize_matrix: Resize an already allocated or reallocate a libscientific
                  matrix
    """
    lsci.ResizeMatrix(mtx, nrows, ncols)


lsci.DelMatrix.argtypes = [ctypes.POINTER(ctypes.POINTER(MATRIX))]
lsci.DelMatrix.restype = None


def del_matrix(mtx):
    """
    del_matrix: Delete an allocated libscientific matrix
    """
    lsci.DelMatrix(ctypes.pointer(mtx))


lsci.MatrixCheck.argtypes = [ctypes.POINTER(MATRIX)]
lsci.MatrixCheck.restype = None


def matrix_check(mtx):
    """
    matrix_check: Find infinite and nan numbers and sobstitute
                 with MISSING value.
    """
    lsci.MatrixCheck(mtx)


lsci.PrintMatrix.argtypes = [ctypes.POINTER(MATRIX)]
lsci.PrintMatrix.restype = None


def print_matrix(mtx):
    """
    print_matrix: Print to video a libscientific matrix
    """
    lsci.PrintMatrix(mtx)


lsci.ValInMatrix.argtypes = [ctypes.POINTER(MATRIX), ctypes.c_double]
lsci.ValInMatrix.restype = ctypes.c_int


def val_in_matrix(mtx, val):
    """
    val_in_matrix: Check if a libscientific matrix contains an exact value "val"
                   and return 1 or 0 respectivelly for yes or no.
    """
    return lsci.ValInMatrix(mtx, val)


lsci.MatrixSet.argtypes = [ctypes.POINTER(MATRIX), ctypes.c_double]
lsci.MatrixSet.restype = None


def matrix_set(mtx, val):
    """
    matrix_set: Set all values of a libscientific matrix to "val"
    """
    lsci.MatrixSet(mtx, val)


lsci.MatrixCopy.argtypes = [ctypes.POINTER(MATRIX),
                            ctypes.POINTER(ctypes.POINTER(MATRIX))]
lsci.MatrixCopy.restype = None


def matrix_copy(msrc, mdst):
    """
    matrix_copy: Copy a libscientifi matrix to another allocated one
    """
    lsci.MatrixCopy(msrc, ctypes.pointer(mdst))


lsci.setMatrixValue.argtypes = [ctypes.POINTER(MATRIX),
                                ctypes.c_size_t,
                                ctypes.c_size_t,
                                ctypes.c_double]
lsci.setMatrixValue.restype = None


def set_matrix_value(mtx, irow, jcol, value):
    """
    set_matrix_value: Set/modify a value in the irow and jcol of a libscientific
                    matrix
    """
    lsci.setMatrixValue(mtx, irow, jcol, value)


lsci.getMatrixValue.argtypes = [ctypes.POINTER(MATRIX),
                                ctypes.c_size_t,
                                ctypes.c_size_t]
lsci.getMatrixValue.restype = ctypes.c_double


def get_matrix_value(mtx, irow, jcol):
    """
    get_matrix_value: Get a value in the irow and jcol of a libscientific
                    matrix
    """
    return lsci.getMatrixValue(mtx, irow, jcol)


def get_matrix_row(mtx, irow):
    """
    get_matrix_row: Python version of getting a row from a libscientific matrix
    """
    row_lst = []
    for j in range(mtx[0].col):
        row_lst.append(get_matrix_value(mtx, irow, j))
    return row_lst


def get_matrix_column(mtx, jcol):
    """
    getMatrixColumn: Python version of getting a column from a libscientific
                     matrix
    """
    col_lst = []
    for i in range(mtx[0].row):
        col_lst.append(get_matrix_value(mtx, i, jcol))
    return col_lst


def matrix_to_list(mtx):
    """
    matrix_to_list: Convert a libscientific matrix to list of list
    """
    mtx_list = []
    try:
        for i in range(mtx[0].row):
            row_lst = []
            for j in range(mtx[0].col):
                row_lst.append(mtx[0].data[i][j])
            mtx_list.append(row_lst)
    except TypeError:
        for i in range(mtx.row):
            row_lst = []
            for j in range(mtx.col):
                row_lst.append(mtx.data[i][j])
            mtx_list.append(row_lst)
    return mtx_list


def matrix_from_numpy(npm):
    """
    Convert a numpy matrix into a libscientific matrix
    """
    return new_matrix(npm.tolist())


def set_missing_matrix_value(mtx, row_id, col_id):
    """
    Set a value of a matrix into a missing value
    """
    mtx[0].data[row_id][col_id] = misc.missing_value()


lsci.MatrixAppendRow.argtypes = [ctypes.POINTER(MATRIX),
                                 ctypes.POINTER(vector.DVECTOR)]
lsci.MatrixAppendRow.restype = None

def matrix_append_row(mtx, row_vect):
    """
    Append to a libscientific matrix a row list
    """
    return lsci.MatrixAppendRow(mtx, row_vect.dvect)


lsci.MatrixAppendCol.argtypes = [ctypes.POINTER(MATRIX),
                                 ctypes.POINTER(vector.DVECTOR)]
lsci.MatrixAppendCol.restype = None

def matrix_append_col(mtx, col_vect):
    """
    Append to a libscientific matrix a column list
    """
    return lsci.MatrixAppendCol(mtx, col_vect.dvect)


lsci.MatrixDVectorDotProduct.argtypes = [ctypes.POINTER(MATRIX),
                                         ctypes.POINTER(vector.DVECTOR),
                                         ctypes.POINTER(vector.DVECTOR)]
lsci.MatrixDVectorDotProduct.restype = None

def matrix_dvector_dot_product(mtx, dvect, res):
    """
    /* Description:
    * matrix - row double vector product: the result is a row double vector
    * i.e.: X(10x5) * d(5x1) = r(10x1)
    */
    void matrix_dvector_dot_product(matrix *mtx, dvector *v, dvector *r);
    """
    return lsci.MatrixDVectorDotProduct(mtx, dvect, res)


lsci.MT_MatrixDVectorDotProduct.argtypes = [ctypes.POINTER(MATRIX),
                                            ctypes.POINTER(vector.DVECTOR),
                                            ctypes.POINTER(vector.DVECTOR)]
lsci.MT_MatrixDVectorDotProduct.restype = None


def mt_matrix_dvector_dot_product(mtx, dvect, res):
    """
    /* Multithread version of matrix_dvector_dot_product */
    void MT_MatrixDVectorDotProduct(matrix *mx, dvector *v, dvector *p);
    """
    return lsci.MT_MatrixDVectorDotProduct(mtx, dvect, res)


lsci.DVectorMatrixDotProduct.argtypes = [ctypes.POINTER(MATRIX),
                                         ctypes.POINTER(vector.DVECTOR),
                                         ctypes.POINTER(vector.DVECTOR)]
lsci.DVectorMatrixDotProduct.restype = None

def dvector_matrix_dot_product(mtx, dvect, res):
    """
    /* Description:
    * column double vector - matrix product: the result is a column double vector
    * i.e.: d(1x5) * X(5x10) = r(1x10)
    */
    void DVectorMatrixDotProduct(matrix *mx, dvector *v, dvector *p);
    """
    return lsci.DVectorMatrixDotProduct(mtx, dvect, res)



lsci.MT_DVectorMatrixDotProduct.argtypes = [ctypes.POINTER(MATRIX),
                                            ctypes.POINTER(vector.DVECTOR),
                                            ctypes.POINTER(vector.DVECTOR)]
lsci.MT_DVectorMatrixDotProduct.restype = None

def mt_dvector_matrix_dot_product(mtx, dvect, res):
    """
    /* Multithread version of DVectorMatrixDotProduct */
    void MT_DVectorMatrixDotProduct(matrix *mx, dvector *v, dvector *p);
    """
    return lsci.MT_DVectorMatrixDotProduct(mtx, dvect, res)


lsci.DVectorTrasposedDVectorDotProduct.argtypes = [ctypes.POINTER(vector.DVECTOR),
                                                   ctypes.POINTER(vector.DVECTOR),
                                                   ctypes.POINTER(MATRIX)]
lsci.DVectorTrasposedDVectorDotProduct.restype = None

def dvector_transposed_dvector_dot_product(mtx, dvect, res):
    """
    /* Description:
    * transposed double vector - double vecrtor product: the result is a matrix
    * i.e.: X  = d'd
    * i.e.: d'(5x1) * d(1x5) = X(1x10)
    */
    void DVectorTrasposedDVectorDotProduct(dvector *v1, dvector *v2, matrix *m);
    """
    return lsci.DVectorTrasposedDVectorDotProduct(mtx, dvect, res)


lsci.DVectorTransposedMatrixDivision.argtypes = [ctypes.POINTER(vector.DVECTOR),
                                                 ctypes.POINTER(MATRIX),
                                                 ctypes.POINTER(vector.DVECTOR)]
lsci.DVectorTransposedMatrixDivision.restype = None

def dvector_transposed_matrix_division(dvect, mtx, res):
    """
    /* r = v/mx  = (inv(mx^T)*v^T)^T*/
    void DVectorTransposedMatrixDivision(dvector *v, matrix *mx, dvector *r);
    """
    return lsci.DVectorTransposedMatrixDivision(dvect, mtx, res)


lsci.MatrixDotProduct.argtypes = [ctypes.POINTER(MATRIX),
                                  ctypes.POINTER(MATRIX),
                                  ctypes.POINTER(MATRIX)]
lsci.MatrixDotProduct.restype = None

def matrix_dot_product(mtx_t, mtx, res):
    """
    /*
    * Description:
    * Calculate the matrix matrix product
    */
    void MatrixDotProduct(matrix *m_t, matrix *mtx, matrix *r);
    """
    return lsci.MatrixDotProduct(mtx_t, mtx, res)


lsci.RowColOuterProduct.argtypes = [ctypes.POINTER(vector.DVECTOR),
                                    ctypes.POINTER(vector.DVECTOR),
                                    ctypes.POINTER(MATRIX)]
lsci.RowColOuterProduct.restype = None

def row_col_outer_product(dvect_a, dvect_b, res):
    """
    /*
    * Description:
    * Calculate the matrix matrix product
    */
    void RowColOuterProduct(dvector *a, dvector *b, matrix *m);
    """
    return lsci.RowColOuterProduct(dvect_a, dvect_b, res)


lsci.MatrixTranspose.argtypes = [ctypes.POINTER(MATRIX),
                                 ctypes.POINTER(MATRIX)]
lsci.MatrixTranspose.restype = None

def matrix_transpose(mtx, mtx_t):
    """
    /*
    * Description:
    * Generate a transpose matrix of m
    */
    void MatrixTranspose(matrix *mtx, matrix *r);
    """
    return lsci.MatrixTranspose(mtx, mtx_t)



lsci.MatrixInversion.argtypes = [ctypes.POINTER(MATRIX),
                                 ctypes.POINTER(MATRIX)]
lsci.MatrixInversion.restype = None

def matrix_inversion(mtx, mtx_inv):
    """
    /*
    * Description:
    * Matrix inversion using the Gauss-Jordan algorithm
    */
    void MatrixInversion(matrix *mtx, matrix **m_inv);
    """
    return lsci.MatrixInversion(mtx, mtx_inv)


lsci.EVectEval.argtypes = [ctypes.POINTER(MATRIX),
                           ctypes.POINTER(vector.DVECTOR),
                           ctypes.POINTER(MATRIX)]
lsci.EVectEval.restype = None

def evect_eval(mtx, mtx_e_vect, mtx_e_val):
    """
    Eigenvectors and Eigenvalues using a lapack method
    """
    return lsci.EVectEval(mtx, mtx_e_vect, mtx_e_val)


lsci.SVD.argtypes = [ctypes.POINTER(MATRIX),
                     ctypes.POINTER(MATRIX),
                     ctypes.POINTER(MATRIX),
                     ctypes.POINTER(MATRIX)]
lsci.SVD.restype = None

def svd(mtx, u_mtx, s_mtx, vt_mtx):
    """
    Calculate the SVD of a matrix mtx using the eigenvector/values lapack method
    """
    lsci.SVD(mtx, u_mtx, s_mtx, vt_mtx)


lsci.SVDlapack.argtypes = [ctypes.POINTER(MATRIX),
                           ctypes.POINTER(MATRIX),
                           ctypes.POINTER(MATRIX),
                           ctypes.POINTER(MATRIX)]
lsci.SVDlapack.restype = None

def svd_lapack(mtx, u_mtx, s_mtx, vt_mtx):
    """
    Calculate the SVD of a matrix mtx using the SVD lapack method
    """
    lsci.SVDlapack(mtx, u_mtx, s_mtx, vt_mtx)


class Matrix():
    """
    Translate a list of list into a libscientific matrix
    """
    def __init__(self, mx_):
        self.mtx = new_matrix(mx_)

    def __del__(self):
        del_matrix(self.mtx)
        del self.mtx
        self.mtx = None

    def __getitem__(self, keys):
        i, j = keys
        return self.data_ptr()[i][j]

    def __setitem__(self, keys, value):
        i, j = keys
        set_matrix_value(self.mtx, i, j, value)

    def nrow(self):
        """
        Return the number of rows of the matrix
        """
        return self.mtx[0].row

    def ncol(self):
        """
        Return the number of columns of the matrix
        """
        return self.mtx[0].col

    def data_ptr(self):
        """
        Return the matrix data pointer
        """
        return self.mtx[0].data

    def tolist(self):
        """
        Return the matrix into a list form
        """
        return matrix_to_list(self.mtx)

    def fromlist(self, mtx_):
        """
        Set the matrix from a list
        """
        del_matrix(self.mtx)
        del self.mtx
        self.mtx = new_matrix(mtx_)

    def fromnumpy(self, npmx):
        """
        Set the matrix from a numpy array
        """
        del_matrix(self.mtx)
        del self.mtx
        self.mtx = matrix_from_numpy(npmx)

    def appendrow(self, row_lst_):
        """
        Append a row to the matrix
        """
        row_lst = vector.DVector(row_lst_)
        matrix_append_row(self.mtx, row_lst)
        del row_lst

    def appendcol(self, col_lst_):
        """
        Append a columnt vector to the matrix
        """
        col_lst = vector.DVector(col_lst_)
        matrix_append_col(self.mtx, col_lst)
        del col_lst

    def transpose(self,):
        """
        Transpose the matrix
        """
        mtx_t = init_matrix()
        resize_matrix(mtx_t, self.mtx[0].col, self.mtx[0].row)
        matrix_transpose(self.mtx, mtx_t)
        matrix_copy(mtx_t, self.mtx)
        del_matrix(mtx_t)
        del mtx_t

    def get_evect_evals(self):
        """
        Return the eigenvectors and eigenvalues of the matrix
        """
        mtx_e_vect_ = vector.init_dvector()
        mtx_e_vals_ = init_matrix()
        evect_eval(self.mtx,  mtx_e_vect_, mtx_e_vals_)
        mtx_e_vect = vector.dvector_tolist( mtx_e_vect_)
        mtx_e_vals = matrix_to_list( mtx_e_vals_)
        del_matrix( mtx_e_vals_)
        vector.del_dvector( mtx_e_vect_)
        return  mtx_e_vect,  mtx_e_vals

    def svd(self):
        """
        Return the SVD U,S and VT of the matrix
        """
        mtx_u = init_matrix()
        mtx_s = init_matrix()
        mtx_vt = init_matrix()
        svd_lapack(self.mtx, mtx_u, mtx_s, mtx_vt)
        mtx_u_lst = matrix_to_list(mtx_u)
        mtx_s_lst = matrix_to_list(mtx_s)
        mtx_vt_lst = matrix_to_list(mtx_vt)
        del_matrix(mtx_u)
        del_matrix(mtx_s)
        del_matrix(mtx_vt)
        return mtx_u_lst, mtx_s_lst, mtx_vt_lst

    def debug(self):
        """
        Debug the matrix content
        """
        print_matrix(self.mtx)


if __name__ in "__main__":
    from random import random
    from random import seed
    seed(123)
    print("Create a matrix of 10x2 of random values")
    print("First prepare a list of list")
    a = [[random() for j in range(2)] for i in range(10)]
    print("Modify the row 4 column 1 with a string'aiuasd'")
    a[4][1] = "aiuasd"

    print("Convert the list of list matrix into matrix type")
    m = Matrix(a)
    m.debug()


    print("Get value at row 1 column 1")
    print(m[1, 1])
    print("Set value at row 1 column 1 with value -2 ")
    m[1, 1] = -2.
    m.debug()
    print("Convert the matrix into a list of list")
    mlst = m.tolist()
    for row in mlst:
        print(row)

    print("Append the row [99, 100]")
    row = [99, 100]
    m.appendrow(row)
    m.debug()

    print("Append the a col [80, 81, 82, 83, 84, 85, 86 ,87, 88, 89, 90]")
    col = [80, 81, 82, 83, 84, 85, 86 ,87, 88, 89, 90]
    m.appendcol(col)
    m.debug()

    print("Test matrix_dvector_dot_product")
    d = vector.DVector([1, 2, 3])
    r = vector.DVector([0 for i in range(11)])
    matrix_dvector_dot_product(m.mtx, d.dvect, r.dvect)
    r.debug()
    del r
    r = vector.DVector([0 for i in range(11)])
    print("Test MT_MatrixDVectorDotProduct")
    mt_matrix_dvector_dot_product(m.mtx, d.dvect, r.dvect)
    r.debug()
    del d
    del r

    print("Test DVectorMatrixDotProduct")
    d = vector.DVector(list(range(1, 12)))
    r = vector.DVector([0 for i in range(3)])
    dvector_matrix_dot_product(m.mtx, d.dvect, r.dvect)
    r.debug()
    del r
    r = vector.DVector([0 for i in range(3)])
    print("Test MT_DVectorMatrixDotProduct")
    mt_dvector_matrix_dot_product(m.mtx, d.dvect, r.dvect)
    r.debug()
    del d
    del r
    del m

    print("Test DVectorTrasposedDVectorDotProduct")
    d1 = vector.DVector(list(range(10)))
    d2 = vector.DVector(list(range(10)))
    m = Matrix([[0 for j in range(10)] for i in range(10)])
    dvector_transposed_dvector_dot_product(d1.dvect, d2.dvect, m.mtx)
    m.debug()
    del d1
    del d2
    del m

    print("Test DVectorTransposedMatrixDivision")
    d = vector.DVector([random() for i in range(10)])
    m = Matrix([[random() for j in range(10)] for i in range(10)])
    r = vector.DVector([0 for i in range(10)])
    dvector_transposed_matrix_division(d.dvect, m.mtx, r.dvect)
    r.debug()
    del d
    del r
    del m


    print("Test MatrixDotProduct")
    m_t = Matrix([[random() for j in range(10)] for i in range(3)])
    m = Matrix([[random() for j in range(3)] for i in range(10)])
    r = Matrix([[0 for j in range(3)] for i in range(3)])
    matrix_dot_product(m_t.mtx, m.mtx, r.mtx)
    r.debug()
    del m_t
    del m
    del r

    print("Test RowColOuterProduct == DVectorTrasposedDVectorDotProduct")
    a = vector.DVector(list(range(10)))
    b = vector.DVector(list(range(10)))
    r = Matrix([[0 for j in range(10)] for i in range(10)])
    row_col_outer_product(a.dvect, b.dvect, r.mtx)
    r.debug()
    del a
    del b
    del r

    print("Test MatrixTranspose")
    m = Matrix([[random() for j in range(3)] for i in range(10)])
    r = Matrix([[0 for j in range(10)] for i in range(3)])
    matrix_transpose(m.mtx, r.mtx)
    m.debug()
    r.debug()

    print("Test MatrixTranspose class inner method")
    m.transpose()
    m.debug()
    m.transpose()

    del m
    del r

    print("Test MatrixInversion")
    m = Matrix([[random() for j in range(4)] for i in range(4)])
    r = Matrix([])
    m.debug()
    matrix_inversion(m.mtx, r.mtx)
    r.debug()
    del r
    del m

    print("Test Eigen vectors/values")
    m = Matrix([[random() for j in range(10)] for i in range(10)])
    e_vect = vector.DVector([])
    e_vals = Matrix([])
    m.debug()
    evect_eval(m.mtx, e_vect.dvect, e_vals.mtx)
    e_vect.debug()
    e_vals.debug()
    del e_vect
    del e_vals
    print("Test Eigen vectors/values inner method")
    e_vect, e_vals = m.get_evect_evals()
    for row in e_vect:
        print(row)
    for row in e_vals:
        print(row)

    u, s, vt = m.svd()

    del m
